/***************************************************************************
 *	cdda_ioc.c - CD-DA  I/O Ԥʤ
 *	
 *		PC/AT System call ݡȳϡ1996/8/4
 *	
 *					Programmed by Nobuyuki Maruichi
 *	Note.
 *			SHIFT JIS Kanji Code / JIS X0208-1990
 *			Tab Size 4 / Shift Width = 4
 ***************************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#if defined __GLIBC__ && __GLIBC__ >= 2
# include <ustat.h>
#endif	/* defined __GLIBC__ && __GLIBC__ >= 2 */
#ifdef	TOWNS
# include <linux/towns_cdioc.h>
#else
# include "towns_cdda.h"
#endif	/* TOWNS */

#include "cdda_ioc.h"
#include "cdda_ext.h"

#define	FALSE	0
#define	TRUE	1

/***************************************************************************
 *	ǥХå̻
 *		DEBUG_TOWNS_IO:	
 *		DEBUG_GENERAL:	
 ***************************************************************************/
/*#define	DEBUG_TOWNS_IO	1*/
/*#define	DEBUG_GENERAL 	1*/

char *CdDevice = "/dev/cdrom";
#ifdef	TOWNS
static char cddaIocErrorMessage[80];
static struct cd_toc TocBacklog;
#endif	/* TOWNS */

int TrayUnlocked(void);

/***************************************************************************
 *	Хѿ
 *		cd_fd:
 *		
 ***************************************************************************/
int Cddrv_fd = -1;
char *CddaErrorMessage = "";
const char *machine_name[] = {
	"Reserved", "FM TOWNS (1/2)", "FM TOWNS (1F/2F/1H/2H)", 
	"FM TOWNS II UX", "FM TOWNS (10F/20F/10H/20H)", 
	"FM TOWNS II CX", "FM TOWNS II UG", "FM TOWNS II HR", 
	"FM TOWNS II HG", "FM TOWNS II UR", "Reserved", 
	"FM TOWNS II MA", "FM TOWNS II MX", "FM TOWNS II ME", 
	"Reserved", "FM TOWNS II MF/Fresh", NULL
};

#ifdef TOWNS
# ifdef GENERAL 
cdDigitalAudio townsCDExtdrive = {
	"TOWNS I/O", 
	TownsExtIO,
	TrayUnlocked, 
	townsPlay2, 
	townsPause, 
	townsResume, 
	townsStop, 
	townsStatus, 
	generalStatus, 
	townsToc, 
	townsDoor
};
# endif /* GENERAL*/
cdDigitalAudio townsCDDrive = {
	"TOWNS I/O", 
	TownsExtIO,
	TrayUnlocked, 
	townsPlay2, 
	townsPause, 
	townsResume, 
	townsStop, 
	townsStatus, 
	NULL, 
	townsToc, 
	townsDoor
};
#endif /*TOWNS*/

#ifdef GENERAL 
# if GENERAL == 0
cdDigitalAudio generalCDDrive = {
	"General I/O", 
	GeneralIO,
	NULL, 
	generalPlay2, 
	generalPause, 
	generalResume, 
	generalStop, 
	generalTownsStatus, 
	generalStatus, 
	generalToc, 
	generalDoor
};
# endif /* GENERAL*/
# if GENERAL > 0
cdDigitalAudio generalCDExtdrive = {
	"General I/O", 
	GeneralIO,
	NULL, 
	generalPlay2, 
	generalPause, 
	generalResume, 
	generalStop, 
	generalTownsStatus, 
	generalStatus, 
	generalToc, 
	generalDoor
};
# endif /* GENERAL*/
#endif /* GENERAL*/

cdDigitalAudio *CddaIoTypes[] = {
#if defined(TOWNS)
	&townsCDDrive,
#else
	NULL, 
#endif /*defined(TOWNS)*/
#if defined(TOWNS) && defined(GENERAL)
	&townsCDExtdrive,
#else
	NULL, 
#endif /*defined(TOWNS) && defined(GENERAL)*/
#ifdef GENERAL 
	&generalCDExtdrive,
#else
	NULL,
#endif /* GENERAL*/
};

#ifdef TOWNS
# ifdef GENERAL 
ioType currentIoType = TownsExtIO;
cdDigitalAudio *currentCDDA = &townsCDExtdrive;
# else
ioType currentIoType = TownsIO;
cdDigitalAudio *currentCDDA = &townsCDDrive;
# endif /* GENERAL*/
#else
# ifdef GENERAL 
ioType currentIoType = GeneralIO;
cdDigitalAudio *currentCDDA = &generalCDExtdrive;
# else
#  error cannot make mwxcd.(because cd isnot supported.)
# endif /* GENERAL*/
#endif /*TOWNS*/

/***************************************************************************
 *	CdDriveOpen: CD-ROM Drive (/dev/cdromʤ) 򥪡ץ󤹤
 *	
 *		int CdDriveOpen(void);
 *	
 *		void;
 *	֤
 *		
 *	
 *		
 ***************************************************************************/
int CdDriveOpen(void)
{
	if(Cddrv_fd == -1) {
		Cddrv_fd = open(CdDevice, O_RDONLY);
	}
	return(Cddrv_fd);
}

/***************************************************************************
 *	CdDriveClose: CD-ROM Drive (/dev/cdromʤ) 򥯥
 *	
 *		int CdDriveClose(void);
 *	
 *		void;
 *	֤
 *		
 *	
 *		
 ***************************************************************************/
int CdDriveClose(void)
{
	int status = -1;
	
	if(Cddrv_fd != -1) {
		status = close(Cddrv_fd);
		Cddrv_fd = -1;
	}
	return(status);
}

/***************************************************************************
 *	isMounted: ꤵ줿ǥХޥȤƤ뤫ɤȽꤹ
 *	
 *		int isMounted(int fd);
 *	
 *		int fd;		եǥץʥǥХ
 *	֤
 *		EOF: 		顼ȽǤʤ
 *		NotMounted:	ꤵ줿ǥХϡޥȤƤʤ
 *		NowMounted:	ꤵ줿ǥХϡޥȤƤ
 *	
 *		
 ***************************************************************************/
int isMounted(int fd)
{
	int status = NotMounted;
	struct stat stbuf;
	struct ustat ust;

	if(fstat(fd, &stbuf) != 0)
		return(EOF);
		
	/* Is this a mounted filesystem? */
	if (ustat(stbuf.st_rdev, &ust) == 0) {
		status = NowMounted;
	}
	return(status);
}

/***************************************************************************
 *	TrayUnlocked: CD-DA ˼ưå
 *	
 *		int TrayUnlocked(void);
 *	
 *		void;
 *	֤
 *		
 *	
 *		TOWNS ΥƥॳѤơCD ȥ쥤Υåޤ
 *		
 ***************************************************************************/
int TrayUnlocked(void)
{
#if	defined(CDIOCDOOR) && defined(CDIOCDOOR_UNLOCK)
	if(Cddrv_fd == -1) {
		if (CdDriveOpen() < 0) {
			return(EOF);
		}
	}
	if(isMounted(Cddrv_fd) == NotMounted) {
		ioctl(Cddrv_fd, CDIOCDOOR, CDIOCDOOR_UNLOCK);
	}
#endif	/*defined(CDIOCDOOR) && defined(CDIOCDOOR_UNLOCK)*/

	return(0);
}

/***************************************************************************
 *	track_check: 
 *	
 *		int track_check(struct cd_toc *toc, int track, int endtrack);
 *	
 *		struct cd_toc *toc;		CD-ROM/CD-DA
 *		int track;
 *		int endtrack;
 *	֤
 *		
 *	
 *		
 ***************************************************************************/
int track_check(struct cd_toc *toc, int track, int endtrack)
{
	int counter;
	int c_track;
	
	c_track = track;
	for(counter = track - toc->s_track;
			counter < endtrack + toc->s_track;++counter) {
		if(toc->track[counter].minute & 0x80) {
			break;
		} else {
			c_track = counter + toc->s_track;
		}
	}
	return(c_track);
}

/***************************************************************************
 *	time2frame: ãĤλ־ե졼ɽ
 *	
 *		u_int time2frame(struct cd_time *msf);
 *	
 *		struct cd_time *msf;	ãĤλ־
 *	֤
 *		ե졼
 *	
 *		
 ***************************************************************************/
u_int time2frame(const struct cd_time *msf)
{
	return (((msf->minute & 0x7f) * 60 + msf->second) * 75 + msf->frame);
}

/***************************************************************************
 *	frame2time: ե졼ɽ줿־ cd_time ¤Τˤ
 *	
 *		void frame2time(u_int t, struct cd_time *msf);
 *	
 *		u_int t;				ե졼ɽ줿־
 *		struct cd_time *msf;	cd_time ¤
 *	֤
 *		
 *	
 *		
 ***************************************************************************/
void frame2time(u_int t, struct cd_time *msf)
{
	msf->minute = t / (60 * 75);
	t %= (60 * 75);
	msf->second = t / 75;
	msf->frame = t % 75;
}

/***************************************************************************
 *	isAudioTrack: ǥȥå
 *	
 *		int isAudioTrack(const struct cd_time *track_ptr);
 *	
 *		const struct cd_time *track_ptr;
 *	֤
 *		:		ǥȥå
 *		:		ǡȥå
 *	
 *		
 ***************************************************************************/
int isAudioTrack(const struct cd_time *track_ptr)
{
	return((track_ptr->minute & 0x80)?(FALSE):(TRUE));
}

/***************************************************************************
 *	isdDataTrack: ǡȥå
 *	
 *		int isDataTrack(const struct cd_time *track_ptr);
 *	
 *		const struct cd_time *track_ptr;
 *	֤
 *		:		ǥȥå
 *		:		ǡȥå
 *	
 *		
 ***************************************************************************/
int isDataTrack(const struct cd_time *track_ptr)
{
	return((track_ptr->minute & 0x80)?(TRUE):(FALSE));
}

#ifdef	TOWNS
/***************************************************************************
 *	FM TOWNS ¸ΥƥॳѤؿ
 *	ؿꥹ
 *		int townsPlay2(struct cdda_play *info);
 *		int townsPause(void);
 *		int townsResume(void);
 *		int townsStop(void);
 *		int townsStatus(struct cd_subq *subq);
 *		int townsToc(struct cd_toc *toc);
 *		int townsDoor(int cmd);
 ***************************************************************************/

/***************************************************************************
 *	townsPlay2: ѹ¤Τ˽äƱդ롣
 *	
 *		int townsPlay2(struct cdda_play *info);
 *	
 *		struct cdda_play *info;		ѹ¤Ρʺ
 *	֤
 *		ţϣơ	顼
 *				ｪλ
 *	
 *		
 ***************************************************************************/
int townsPlay2(struct cdda_play *info)
{
	int status = EOF;

	if (CdDriveOpen() < 0) {
		perror(CdDevice);
		CddaErrorMessage = CdDevice;
		return(EOF);
	}
	if(SubCdTime(&info->end, &info->start) > 0) {
		if (ioctl(Cddrv_fd, CDIOCPLAY, info) < 0) {
			sprintf(cddaIocErrorMessage, "%.40s CDIOCPLAY", CdDevice);
			CddaErrorMessage = cddaIocErrorMessage;
		} else {
			status = 0;
		}
	} else {
		/*printf("Programming Error: Check Source Program!!\n");*/
	}
	return(status);
}

/***************************************************************************
 *	townsPause: ãĤߤ
 *	
 *		int townsPause(void);
 *	
 *		
 *	֤
 *		EOF		顼
 *		0		ｪλ
 *	
 *		
 ***************************************************************************/
int townsPause(void)
{
	int status = 0;
	
	if (CdDriveOpen() < 0){
		CddaErrorMessage = CdDevice;
		return(EOF);
	}
	if (ioctl(Cddrv_fd, CDIOCPAUSE, NULL) < 0) {
	    sprintf(cddaIocErrorMessage, "%.40s CDIOCPAUSE", CdDevice);
		CddaErrorMessage = cddaIocErrorMessage;
		status = EOF;
	}
	/*********************
	 *	close(Cddrv_fd);
	 *********************/
	return(status);
}

/***************************************************************************
 *	townsResume: CD հߤβ
 *	
 *		int townsResume(void);
 *	
 *		
 *	֤
 *		EOF		顼
 *		0		ｪλ
 *	
 *		
 ***************************************************************************/
int townsResume(void)
{
	int status = 0;
	
	if(CdDriveOpen() < 0) {
		CddaErrorMessage = CdDevice;
		return(EOF);
	}
	if (ioctl(Cddrv_fd, CDIOCRESUME, NULL) < 0) {
		sprintf(cddaIocErrorMessage, "%.40s CDIOCRESUME", CdDevice);
		CddaErrorMessage = cddaIocErrorMessage;
		status = EOF;
	}
	/*********************
	 *	close(Cddrv_fd);
	 *********************/
	return(status);
}

/***************************************************************************
 *	townsStop: CD դ
 *	
 *		int townsStop(void);
 *	
 *		
 *	֤
 *		EOF		顼
 *		0		ｪλ
 *	
 *		TOWNS Ѥ ioctl() Ѥ CD ߤ롣߸ˡå
 *		ơCD ǥХ򥯥롣
 *		
 ***************************************************************************/
int townsStop(void)
{
	int status = 0;
	
	if (CdDriveOpen() < 0){
		CddaErrorMessage = CdDevice;
		return(EOF);
	}
	if (ioctl(Cddrv_fd, CDIOCSTOP, NULL) < 0) {
		sprintf(cddaIocErrorMessage, "%.40s CDIOCSTOP", CdDevice);
		CddaErrorMessage = cddaIocErrorMessage;
		status = EOF;
	}
	TrayUnlocked();
	CdDriveClose();
	return(status);
}

/***************************************************************************
 *	townsStatus: CD αվ֤μ
 *	
 *		int townsStatus(struct cd_subq *subq);
 *	
 *		struct cd_subq *subq;	
 *	֤
 *		EOF:	顼
 *		0:		ｪλ
 *	
 *	
 ***************************************************************************/
int townsStatus(struct cd_subq *subq)
{
	if (CdDriveOpen() < 0){
		CddaErrorMessage = CdDevice;
		return(EOF);
	}
	
	subq->playing = 1;
	if (ioctl(Cddrv_fd, CDIOCSUBQ, subq) < 0){
		/*perror("/dev/cdrom CDIOCSUBQ");*/
		CddaErrorMessage = "";
		return(EOF);
	}
	if(subq->playing && subq->track == 0) {
		cd_stop();
		subq->playing = 0x00;
		CddaErrorMessage = "";
		return(EOF);
	}
#ifdef	DEBUG_TOWNS_IO
	if (subq->playing)
		printf("Now, Playing.\n");
	else
		printf("Not Playing.\n");
#if	defined(DEBUG_TOWNS_IO) && DEBUG_TOWNS_IO > 10
	printf("0x%02x\n", subq->reserv1);
#endif	/* defined(DEBUG_TOWNS_IO) && DEBUG_TOWNS_IO > 10 */
	printf("track = %d\n", subq->track);
#if	defined(DEBUG_TOWNS_IO) && DEBUG_TOWNS_IO > 10
	printf("0x%02x\n", subq->reserv2);
#endif	/* defined(DEBUG_TOWNS_IO) && DEBUG_TOWNS_IO > 10 */
	printf("time in track = (%d,%d,%d)\n", 
			subq->offset.minute, subq->offset.second, subq->offset.frame);
	printf("0x%02x\n", subq->reserv3);
	printf("time = (%d,%d,%d)\n", 
			subq->time.minute, subq->time.second, subq->time.frame);
#endif	/*DEBUG_TOWNS_IO*/
	return(0);
}

/***************************************************************************
 *	townsToc: CD TOC:Table of Contents-CDˤɤ߼
 *	
 *		int townsToc(struct cd_toc *toc);
 *	
 *		struct cd_toc *toc;		ãľؤΥݥ
 *	֤
 *		ţϣơ	顼
 *		ʾ塧	ȥå
 *	
 *		
 ***************************************************************************/
int townsToc(struct cd_toc *toc)
{
	int number_of_track;
	
	if (CdDriveOpen() < 0){
		CddaErrorMessage = CdDevice;
		return(EOF);
	}
	if (ioctl(Cddrv_fd, CDIOCTOC, toc) < 0){
		sprintf(cddaIocErrorMessage, "%.40s CDIOCTOC", CdDevice);
		CddaErrorMessage = cddaIocErrorMessage;
		return(EOF);
	} else {
		TocBacklog = *toc;
	}
	number_of_track = toc->e_track - toc->s_track + 1;
	return(number_of_track);
}

#ifdef	CDIOCDOOR
/***************************************************************************
 *	townsDoor: ãĥȥ쥤ĤԤ
 *	
 *		int townsDoor(int cmd);
 *	
 *		int cmd;		Ĥʤɤ򼨤
 *	֤
 *		EOF		顼
 *		0		ｪλ
 *	
 *		
 ***************************************************************************/
int townsDoor(int cmd)
{
	int status = 0;

	if (CdDriveOpen() < 0){
		CddaErrorMessage = CdDevice;
		status = EOF;
	} else {
		if (ioctl(Cddrv_fd, CDIOCDOOR, cmd) < 0) {
			sprintf(cddaIocErrorMessage, "%.40s door ioctl", CdDevice);
			CddaErrorMessage = cddaIocErrorMessage;
			status = EOF;
		}
	}
	CdDriveClose();
	return(status);
}
#endif	/*CDIOCDOOR*/
#endif	/*TOWNS*/

#ifdef	GENERAL
/***************************************************************************
 *	Linux ɸΥƥॳѤؿ
 *	ؿꥹ
 *		int generalPlay2(struct cdda_play *info);
 *		int generalPause(void);
 *		int generalResume(void);
 *		int generalStop(void);
 *		int generalStatus(struct cd_subq *subq);
 *		int generalToc(struct cd_toc *toc);
 *		int generalDoor(int cmd);
 ***************************************************************************/

/***************************************************************************
 *	generalPlay2: ѹ¤Τ˽äƱդ롣
 *	
 *		int generalPlay2(struct cdda_play *info);
 *	
 *		struct cdda_play *info;		ѹ¤Ρʺ
 *	֤
 *		EOF:	顼
 *		0:		ｪλ
 *	
 *		
 ***************************************************************************/
int generalPlay2(struct cdda_play *info)
{
	int status = EOF;
	struct cdrom_msf msf;

#ifdef	DEBUG_GENERAL
	printf("%s:%d:generalPlay2();start\n", __FILE__, __LINE__);
#endif	/*DEBUG_GENERAL*/
	msf.cdmsf_min0   = info->start.minute;
	msf.cdmsf_sec0   = info->start.second;
	msf.cdmsf_frame0 = info->start.frame;
	msf.cdmsf_min1   = info->end.minute;
	msf.cdmsf_sec1   = info->end.second;
	msf.cdmsf_frame1 = info->end.frame;

	if(CdDriveOpen() < 0) {
		perror(CdDevice);
		CddaErrorMessage = CdDevice;
		return(EOF);
	}
	if(ioctl(Cddrv_fd, CDROMSTART))
		return(status);
	if(ioctl(Cddrv_fd, CDROMPLAYMSF, &msf) == 0) {
		status = 0;
	}
#ifdef	DEBUG_GENERAL
	printf("%s:%d:generalPlay2();end\n", __FILE__, __LINE__);
#endif	/*DEBUG_GENERAL*/
	return(status);
}

/***************************************************************************
 *	generalPause: ãĤߤ
 *	
 *		int generalPause(void);
 *	
 *		
 *	֤
 *		EOF		顼
 *		0		ｪλ
 *	
 *		
 ***************************************************************************/
int generalPause(void)
{
	int status = 0;

#ifdef	DEBUG_GENERAL
	printf("%s:%d:generalPause();start\n", __FILE__, __LINE__);
#endif	/*DEBUG_GENERAL*/
	if (CdDriveOpen() < 0) {
		perror(CdDevice);
		CddaErrorMessage = CdDevice;
		return(EOF);
	}
	if(ioctl(Cddrv_fd, CDROMPAUSE))
		status = EOF;
	
#ifdef	DEBUG_GENERAL
	printf("%s:%d:generalPause();\n", __FILE__, __LINE__);
#endif	/*DEBUG_GENERAL*/
	return(status);
}

/***************************************************************************
 *	generalResume: CD հߤβ
 *	
 *		int generalResume(void);
 *	
 *		
 *	֤
 *		EOF		顼
 *		0		ｪλ
 *	
 *		
 ***************************************************************************/
int generalResume(void)
{
	int status = 0;

#ifdef	DEBUG_GENERAL
	printf("%s:%d:generalResume();start\n", __FILE__, __LINE__);
#endif	/*DEBUG_GENERAL*/
	if (CdDriveOpen() < 0) {
		perror(CdDevice);
		CddaErrorMessage = CdDevice;
		return(EOF);
	}
	if(ioctl(Cddrv_fd, CDROMRESUME))
		status = EOF;
#ifdef	DEBUG_GENERAL
	printf("%s:%d:generalResume();\n", __FILE__, __LINE__);
#endif	/*DEBUG_GENERAL*/
	
	return(status);
}

/***************************************************************************
 *	generalStop: CD դ
 *	
 *		int generalStop(void);
 *	
 *		
 *	֤
 *		EOF		顼
 *		0		ｪλ
 *	
 *		
 ***************************************************************************/
int generalStop(void)
{
	int status = 0;

#ifdef	DEBUG_GENERAL
	printf("%s:%d:generalStop();start\n", __FILE__, __LINE__);
#endif	/*DEBUG_GENERAL*/
	if (CdDriveOpen() < 0) {
		perror(CdDevice);
		CddaErrorMessage = CdDevice;
		return(EOF);
	}
	if(ioctl(Cddrv_fd, CDROMSTOP))
		status = EOF;
	
#ifdef	DEBUG_GENERAL
	printf("%s:%d:generalStop();\n", __FILE__, __LINE__);
#endif	/*DEBUG_GENERAL*/
	return(status);
}

/***************************************************************************
 *	generalStatus: CD αվ֤μ
 *	
 *		int generalStatus(struct cd_subq *subq);
 *	
 *		struct cd_subq *subq;	
 *	֤
 *		EOF		顼
 *		0		ｪλ
 *	
 *		
 ***************************************************************************/
int generalStatus(struct cd_subq *subq)
{
	struct cdrom_subchnl subChanel;

#ifdef	DEBUG_GENERAL
	printf("%s:%d:generalStatus();start\n", __FILE__, __LINE__);
#endif	/*DEBUG_GENERAL*/
	subq->playing = CS_UNKNOWN;
	if (CdDriveOpen() < 0){
		CddaErrorMessage = CdDevice;
		/*printf("Device not opend!!\n");*/
		return(EOF);
	}
	subChanel.cdsc_format = CDROM_MSF;
	if(ioctl(Cddrv_fd, CDROMSUBCHNL, &subChanel)) {
		subq->playing = CS_EJECTED;
		/*perror("/dev/cdrom CDROMSUBCHNL");*/
		CddaErrorMessage = "";
		return (1);
	}
	subq->track = subChanel.cdsc_trk;
	subq->offset.minute = subChanel.cdsc_reladdr.msf.minute;
	subq->offset.second = subChanel.cdsc_reladdr.msf.second;
	subq->offset.frame  = subChanel.cdsc_reladdr.msf.frame;
	subq->time.minute = subChanel.cdsc_absaddr.msf.minute;
	subq->time.second = subChanel.cdsc_absaddr.msf.second;
	subq->time.frame  = subChanel.cdsc_absaddr.msf.frame;

	switch (subChanel.cdsc_audiostatus) {
	case CDROM_AUDIO_PLAY:
		/*printf("Now playing!!\n");*/
		subq->playing = CS_PLAYING;
		break;
	case CDROM_AUDIO_PAUSED:
	case CDROM_AUDIO_NO_STATUS:
	case CDROM_AUDIO_INVALID: /**/
		/*printf("Stopping or pausing!!\n");*/
		subq->playing = CS_STOP;	/* STOP/PAUSE ȽԲ */
		break;
	case CDROM_AUDIO_COMPLETED:
		/*printf("Waiting!!\n");*/
		subq->playing = CS_WAITING;
		break;
	default:
		/*printf("Unknown!!\n");*/
		subq->playing = CS_UNKNOWN;
		break;
	}
#ifdef	DEBUG_GENERAL
	printf("%s:%d:generalStatus();\n", __FILE__, __LINE__);
#endif	/*DEBUG_GENERAL*/
	return(0);
}

/***************************************************************************
 *	generalTownsStatus: CD αվ֤μ
 *	
 *		int generalTownsStatus(struct cd_subq *subq);
 *	
 *		struct cd_subq *subq;	
 *	֤
 *		EOF		顼
 *		0		ｪλ
 *	
 *		
 ***************************************************************************/
int generalTownsStatus(struct cd_subq *subq)
{
	int status = EOF;

	if(generalStatus(subq) == 0) {
		if(subq->playing == CS_PLAYING) {
			status = 0;
		}
	}
	return(status);
}

/***************************************************************************
 *	generalTocTrack: CD TOC:Table of Contents-CDˤɤ߼
 *	
 *		static int generalTocTrack(int fd, int track, 
 *					struct cd_toc *toc_ptr, struct cd_time *track_ptr);
 *	
 *		int fd;
 *		int track;
 *		struct cd_toc *toc_ptr;		ãľؤΥݥ
 *		struct cd_time *track_ptr;
 *	֤
 *		EOF:	顼
 *		1ʾ:	ȥå
 *	
 *		
 ***************************************************************************/
static int generalTocTrack(int fd, int track, 
						struct cd_toc *toc_ptr, struct cd_time *track_ptr)
{
	int status = 0;
	int dataTrack;
	struct cdrom_tocentry entry;

#if	defined(DEBUG_GENERAL) && DEBUG_GENERAL > 1
	printf("%s:%d:generalTocTrack();:#%d\n", __FILE__, __LINE__, track);
#endif	/* defined(DEBUG_GENERAL) && DEBUG_GENERAL > 1 */
	entry.cdte_track = track;
	entry.cdte_format = CDROM_MSF;
	if (ioctl(fd, CDROMREADTOCENTRY, &entry)) {
		status = EOF;
	} else {
		dataTrack = ((entry.cdte_ctrl & CDROM_DATA_TRACK)?(1):(0));
		track_ptr->minute = 
					(dataTrack?0x80:0x00) | entry.cdte_addr.msf.minute;
		track_ptr->second = entry.cdte_addr.msf.second;
		track_ptr->frame  = entry.cdte_addr.msf.frame;
		if(track != CDROM_LEADOUT) {
			toc_ptr->type |= ((dataTrack)?(CD_ROM_Data):(CD_DigitalAudio));
		}
	}
#if	defined(DEBUG_GENERAL) && DEBUG_GENERAL > 1
	printf("%s:%d:generalTocTrack();\n", __FILE__, __LINE__);
#endif	/* defined(DEBUG_GENERAL) && DEBUG_GENERAL > 1 */
	return(status);
}

/***************************************************************************
 *	generalToc: CD TOC:Table of Contents-CDˤɤ߼
 *	
 *		int generalToc(struct cd_toc *toc);
 *	
 *		struct cd_toc *toc;		ãľؤΥݥ
 *	֤
 *		ţϣơ	顼
 *		ʾ塧	ȥå
 *	
 *		
 ***************************************************************************/
int generalToc(struct cd_toc *toc)
{
	int track;
	int status = 0;
	struct cdrom_tochdr	hdr;
	struct cd_toc tocBuffer, *toc_ptr;

#ifdef	DEBUG_GENERAL
	printf("%s:%d:generalToc();start\n", __FILE__, __LINE__);
#endif	/*DEBUG_GENERAL*/
	toc_ptr = &tocBuffer;
	if (CdDriveOpen() < 0) {
		perror(CdDevice);
		CddaErrorMessage = CdDevice;
		return(EOF);
	}
	if(ioctl(Cddrv_fd, CDROMREADTOCHDR, &hdr)) {
		return(EOF);
	}
	toc_ptr->type = 0;
	toc_ptr->s_track = hdr.cdth_trk0;
	toc_ptr->e_track = hdr.cdth_trk1;
	for(track = toc_ptr->s_track;
					track <= toc_ptr->e_track && status == 0;++track) {
		status = generalTocTrack(Cddrv_fd, track, toc_ptr, 
						&toc_ptr->track[track - toc_ptr->s_track]);
	}
	if(status == 0) {
		status = generalTocTrack(Cddrv_fd, 
							CDROM_LEADOUT, toc_ptr, &toc_ptr->total);
		if(status == 0) {
			*toc = *toc_ptr;
		}
	}
#ifdef	DEBUG_GENERAL
	printf("%s:%d:generalToc();:end\n", __FILE__, __LINE__);
#endif	/*DEBUG_GENERAL*/
	return((status)?(EOF):(toc_ptr->e_track-toc_ptr->s_track+1));
}

/***************************************************************************
 *	generalDoor: ãĥȥ쥤ĤԤ
 *	
 *		int generalDoor(int cmd);
 *	
 *		int cmd;		Ĥʤɤ򼨤
 *	֤
 *		EOF		顼
 *		0		ｪλ
 *	
 *		ɸ ioctl() Ѥ CD ߤ롣ɸ CD ǥХϡ
 *		ץΤȤ˥ȥ쥤Ĥ褦Ǥ
 *		ΤᡢǡCD ǥХ򥯥ȡδؿȤϡ
 *		̤ǰ֤˹Ԥȥ쥤γľ֥åΤȤǥХ
 *		ץ󤹤롢κݤ˥ȥ쥤ĤޤäƤޤΤǡȥ쥤򳫤
 *		̵̣ˤʤäƤޤޤ
 *		εưΰ㤤ΤᡢTOWNS Ѥ ioctl() ǤΥȥ쥤γĤȤϡ
 *		ۤʤޤ
 *		
 ***************************************************************************/
int generalDoor(int cmd)
{
	int status = 0;

	if (CdDriveOpen() < 0){
		CddaErrorMessage = CdDevice;
		status = EOF;
	} else {
		switch(cmd) {
		default:
#ifdef CDIOCDOOR_UNLOCK
		case CDIOCDOOR_UNLOCK:
#endif	/*CDIOCDOOR_UNLOCK*/
#ifdef CDIOCDOOR_LOCK
		case CDIOCDOOR_LOCK:
#endif	/*CDIOCDOOR_LOCK*/
			break;
		case CDIOCDOOR_OPEN:
			if(isMounted(Cddrv_fd) == NotMounted) {
				if(ioctl(Cddrv_fd, CDROMEJECT)) {
					return(EOF);
				}
			}
			break;
		case CDIOCDOOR_CLOSE:
			/* CD ȥ쥤򥯥ʤĤ */
			if(ioctl(Cddrv_fd, CDROMCLOSETRAY)) {
				return(EOF);
			}
			break;
		}
	}
	/* CdDriveClose(); */
	return(status);
}
#endif	/*GENERAL*/

/***************************************************************************
 *	cd_play2: ѹ¤Τ˽äƱդ롣
 *	
 *		int cd_play2(struct cdda_play *info);
 *	
 *		struct cdda_play *info;		ѹ¤Ρʺ
 *	֤
 *		EOF:	顼
 *		0:		ｪλ
 *	
 *		
 ***************************************************************************/
int cd_play2(struct cdda_play *info)
{
	int status = EOF;

	if(currentCDDA->play != NULL)
		status = currentCDDA->play(info);
	return(status);
}

/***************************************************************************
 *	cd_pause: ãĤߤ
 *	
 *		int cd_pause(void);
 *	
 *		
 *	֤
 *		EOF:	顼
 *		0:		ｪλ
 *	
 *		
 ***************************************************************************/
int cd_pause(void)
{
	int status = EOF;

	if(currentCDDA->pause != NULL)
		status = currentCDDA->pause();
	return(status);
}

/***************************************************************************
 *	cd_resume: CD հߤβ
 *	
 *		int cd_resume(void);
 *	
 *		void;
 *	֤
 *		EOF:	顼
 *		0:		ｪλ
 *	
 *		
 ***************************************************************************/
int cd_resume(void)
{
	int status = EOF;

	if(currentCDDA->resume != NULL)
		status = currentCDDA->resume();
	return(status);
}

/***************************************************************************
 *	cd_stop: CD դ
 *	
 *		int cd_stop(void);
 *	
 *		
 *	֤
 *		EOF:	顼
 *		0:		ｪλ
 *	
 *		
 ***************************************************************************/
int cd_stop(void)
{
	int status = EOF;

	if(currentCDDA->stop != NULL)
		status = currentCDDA->stop();
	return(status);
}

/***************************************************************************
 *	cd_status: CD αվ֤μ
 *	
 *		int cd_status(struct cd_subq *subq);
 *	
 *		struct cd_subq *subq;	
 *	֤
 *		EOF:	顼
 *		0:		ｪλ
 *	
 *	
 ***************************************************************************/
int cd_status(struct cd_subq *subq)
{
	int status = EOF;

	if(currentCDDA->status != NULL)
		status = currentCDDA->status(subq);
	return(status);
}

/***************************************************************************
 *	cddaStatus2: CD αվ֤μ
 *	
 *		int cddaStatus2(struct cd_subq *subq);
 *	
 *		struct cd_subq *subq;	
 *	֤
 *		EOF		顼
 *		0		ｪλ
 *	
 *		
 ***************************************************************************/
int cddaStatus2(struct cd_subq *subq)
{
	int status = EOF;

	if(currentCDDA->status2 != NULL)
		status = currentCDDA->status2(subq);
	return(status);
}

/***************************************************************************
 *	get_toc: CD TOC:Table of Contents-CDˤɤ߼
 *	
 *		int get_toc(struct cd_toc *toc);
 *	
 *		struct cd_toc *toc;		ãľؤΥݥ
 *	֤
 *		EOF		顼
 *		ʾ塧	ȥå
 *	
 *		
 ***************************************************************************/
int get_toc(struct cd_toc *toc)
{
	int status = EOF;

	if(currentCDDA->toc != NULL)
		status = currentCDDA->toc(toc);
	return(status);
}

/***************************************************************************
 *	cd_door_cntrl: ãĥȥ쥤ĤԤ
 *	
 *		int cd_door_cntrl(int cmd);
 *	
 *		int cmd;		Ĥʤɤ򼨤
 *	֤
 *		EOF		顼
 *		0		ｪλ
 *	
 *		
 ***************************************************************************/
int cd_door_cntrl(int cmd)
{
	int status = EOF;

	if(currentCDDA->doorControl != NULL)
		status = currentCDDA->doorControl(cmd);
	return(status);
}

/***************************************************************************
 *	get_machine_id: FM ԣϣףΣӤε̤ѡ
 *	
 *		int get_machine_id(void);
 *	֤
 *		EOF			ɣĤǤʤޤϡTOWNS ʳ
 *		ʾ		ɣ
 *	
 *		
 ***************************************************************************/
int get_machine_id(void)
{
#ifdef	TOWNS
	int fd;
	char machine_id[2];
#endif	/*TOWNS*/
	int status = EOF;

#ifdef	TOWNS
	if ((fd=open("/dev/port", O_RDONLY))<0) {
		return(EOF);
	}

	lseek(fd, MACHINE_ID1,0);
	read(fd, &machine_id, 2);
	if((machine_id[0] & 0xf8) == 0x00) {
		status = machine_id[1];
	}
	close(fd);
	
#endif	/*TOWNS*/
	return(status);
}

