#ifndef lint
static char sccsid[] = "%W% %G%";
#endif

#include <stdio.h>

#define CATENATE    "/bin/cat"
#define CMD_SEP     "@"
#define STDIN       0
#define STDOUT      1

char *progname;					/* command name at execution */

/***************************************************************************
 *	pipe_sample: 
 *	`
 *		int pipe_sample(char **catv, char **cmds);
 *	
 *		char **catv;		
 *		char **cmds;		
 *	Ԃl
 *		
 *	
 *		
 ***************************************************************************/
int pipe_sample(char **catv, char **cmds)
{
	int fds[2];					/* for pipe(2) call */
	int nchild;
	int cat_pid, cmd_pid, pid;
	int cat_status, cmd_status, status;
	int cat_reason, cmd_reason;

	if (pipe(fds)) {
		perror("pipe()");		/* File table overflow */
		return (-1);
	}
	if ((cat_pid = fork()) == 0) {
		/* child process, 'cat' execute */
		close(STDOUT);
		dup(fds[1]);
		close(fds[0]);
		close(fds[1]);
		execv(catv[0], catv);
		perror(catv[0]);
		_exit(1);
		/*NOTREACHED */
	}
	if ((cmd_pid = fork()) == 0) {
		/* child process, specified command */
		close(STDIN);
		dup(fds[0]);
		close(fds[0]);
		close(fds[1]);
		execvp(cmds[0], cmds);
		perror(cmds[0]);
		_exit(1);
		/*NOTREACHED */
	}
	if (cat_pid == -1 || cmd_pid == -1) {
		perror("sorry");		/* No more processes */
		goto error;
	}
	/* parent process */
	close(fds[0]);
	close(fds[1]);
	/* wait for children */
	nchild = 2;
	do {
	  retry:
		if ((pid = wait(&status)) == -1) {
			perror("wait()");	/* No children */
			goto error;
		}
		if (pid == cat_pid)
			cat_status = status;
		else if (pid == cmd_pid)
			cmd_status = status;
		else
			goto retry;
	} while (--nchild > 0);
	cmd_reason = analize(cmd_pid, cmd_status);
	cat_reason = analize(cat_pid, cat_status);
	if (cmd_reason != 0)
		return (cmd_reason);
	if (cat_reason != 0)
		return (cat_reason);
	return (0);					/* success */
  error:
	close(fds[0]);
	close(fds[1]);
	return (-1);				/* something failure */
}

/***************************************************************************
 *	analize: 
 *	`
 *		int analize(int pid, int status);
 *	
 *		int pid;		
 *		int status;		
 *	Ԃl
 *		
 *	
 *		
 ***************************************************************************/
int analize(int pid, int status)
{
	int term, reason;

	if (status & 0200) {
		fprintf(stderr, "%d: core dumped\n", pid);
		return (-1);
	}
	if ((term = status & 0377) != 0) {
		fprintf(stderr, "%d: killed by signal(%d)\n", pid, term);
		return (-1);
	}
	if ((reason = (status >> 8) & 0377) != 0) {
		fprintf(stderr, "%d: exit %d\n", pid, reason);
		return (reason);
	}
	return (0);					/* all right */
}

/***************************************************************************
 *	main: 
 *	`
 *		int main(int argc, char **argv);
 *	
 *		int argc;		
 *		char **argv;		
 *	Ԃl
 *		
 *	
 *		
 ***************************************************************************/
int main(int argc, char **argv)
{
	char *malloc();
	char **cmds;
	char **catv;
	int catc;
	int i;
	int status;
	int found;

	progname = argv[0];
	if (argc < 3) {
		fprintf(stderr,
				"usage: %s [<file>]... %s <cmd> [<args>]...\n",
				progname, CMD_SEP);
		exit(1);
	}
	found = 0;
	for (catc = 1; catc < argc; catc++)
		if (strcmp(argv[catc], CMD_SEP) == 0) {
			found = 1;
			break;
		}
	if (!found) {
		fprintf(stderr, "%s: Missing command\n", progname);
		exit(1);
	}
	cmds = &argv[catc + 1];
	if ((catv = (char **) malloc(sizeof(char *) * (catc + 1))) == NULL) {
		perror("sorry");		/* Not enough core */
		exit(1);
	}
	catv[0] = CATENATE;
	for (i = 1; i < catc; i++)
		catv[i] = argv[i];
	catv[catc] = NULL;
	status = pipe_sample(catv, cmds);
	free((char *) catv);
	exit(status);
	/*NOTREACHED */
}
